%{
/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btsrv-parser.y,v 1.16 2002/11/14 10:04:03 kds Exp $

   parser for btsrv configuration file

   Fixes:	
   		Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#include <affix/config.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/errno.h>

#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <getopt.h>
#include <string.h>
#include <termios.h>

#include <affix/bluetooth.h>
#include <affix/btcore.h>

#include "btsrv.h"


struct {
	char		name[MAX_PROFILE_NAME_LEN];
	sdpreg_func_t	reg_func;
} profiles[] =
#ifdef CONFIG_AFFIX_SDP
{
        {"SerialPort", sdpreg_serialport},
        {"DialupNetworking", sdpreg_dialupnetworking},
	{"LANAccess", sdpreg_lanaccess},
        {"OBEXFiletransfer", sdpreg_obex_filetransfer},
	{"OBEXObjectPush", sdpreg_obex_objectpush}
};
#else
{
        {"SerialPort", NULL},
        {"DialupNetworking", NULL},
	{"LANAccess", NULL},
        {"OBEXFiletransfer", NULL},
	{"OBEXObjectPush", NULL}
};
#endif

#define NUM_PROFILES ((int)(sizeof(profiles) / sizeof(profiles[0])))


//#define YYPARSE_PARAM	param
//#define YYLEX_PARAM	param

int yylex(void);
void yyerror(char *s);
void myyyerror(char *format, ...);

extern int yylineno;

static char		*config_file;
static int	 	svcnum;
static int		devnum;
static device_t		*dev;
static service_t	*svc;

int	waitsemi = 0;

%}

/* bison declaration */
//%defines
//%name-prefix="btsrv_"
//%output="btsrv-config.c"

%union {
	int	num;
	char	*str;
}

/* Configuration attribute tokens */
%token	OPT_DEVICE OPT_SERVICE OPT_PROFILE OPT_FLAGS OPT_EXEC OPT_PROVIDER OPT_DESCRIPT
%token	OPT_SECURITY OPT_COD OPT_NAME OPT_ROLE OPT_PKTTYPE OPT_SCANMODE


%token <num>	NUM
%token <str>	STRING WORDLIST WORD

%type <str>	name
%type <str>	string params

%start config

%%

string:	WORD | STRING ;

params:	WORD | WORDLIST ;
	
name:	'*' { $$ = NULL; }
	| string
	;
	
optend:	{ waitsemi = 1; } ';';


config:	/* empty */
	| list
	;

list:	entry
	| list entry
	;

entry:	device	| service
	;


/* ------------------------------------------------------------------- */

device:
	'<' OPT_DEVICE name 
			{
				char	*name = $3;
				if (name == NULL) {
					BTDEBUG("found common device\n");
					if (devnum)
						devices[devnum] = devices[0];
					dev = &devices[0];
					dev->name[0] = '*';
				} else {
					BTDEBUG("found device: %s", name);
					/* get device num */
					dev = &devices[devnum];
					strcpy(dev->name, name);
				}
			}
	'>'
		devopt_list
	'<' '/' OPT_DEVICE '>' { devnum++; }
	;


devopt_list:	/*empty*/
		| devopt_list1
		;

devopt_list1:	devopt optend
		| devopt_list1 devopt optend
		;

devopt:
	OPT_NAME string
			{
				dev->btname = strdup($2);
			}
	| OPT_COD params
			{
				str2cod($2, &dev->cod);
				BTDEBUG("cod: %#.6x", dev->cod);
			}
	| OPT_ROLE params
			{
				str2mask(role_map, $2, &dev->role);
				;
			}
	| OPT_SECURITY params
			{
				str2mask(sec_mode_map, $2, &dev->security);
				BTDEBUG("sec_mode: %#x", dev->security);
			}
	| OPT_SCANMODE params
			{
				str2mask(scan_map, $2, &dev->scan);
				BTDEBUG("scan_mode: %#x", dev->scan);

			}
	| OPT_PKTTYPE params
			{
				str2mask(pkt_type_map, $2, &dev->pkt_type);
				BTDEBUG("pkt_type: %#x", dev->pkt_type);
			}
	//| error { /*yyclearin; yyerrok;*/ }
	;



/* ------------------------------------------------------------------------------ */
service:
	'<' OPT_SERVICE string 
			{ 
				svc = &services[svcnum];
				svc->name = strdup($3);
				BTDEBUG("found service: [%s]", $3);
			} 
		'>'
		svcopt_list
	'<' '/' OPT_SERVICE '>'	{ svcnum++; }
	;

svcopt_list:	/*empty*/
		| svcopt_list1
		;

svcopt_list1:	svcopt optend
		| svcopt_list1 svcopt optend
		;

svcopt:	
	OPT_PROFILE string
			{ 
				int	i;
				BTDEBUG("found profile: %s", $2); 
				for (i = 0; i < NUM_PROFILES; i++)
				if (strcasecmp(profiles[i].name, $2) == 0)
					break;
				if (i == NUM_PROFILES) {
					BTERROR("Invalid profile name %s in config file %s line %d",
						$2, config_file, yylineno);
					return -1;
				}
#if defined(CONFIG_AFFIX_SDP)
				svc->reg_func = profiles[i].reg_func;
#endif
				// process flags
			}
	| OPT_FLAGS params
			{ 
				struct affix_tupla	flags[] = {
						{SRV_FLAG_RFCOMM_TTY, "tty"},
						{SRV_FLAG_STD, "std"},
						{0, "socket"},
						{0, 0}
						};

				BTDEBUG("found flag list: [%s]", $2);
				if (!str2mask(flags, $2, &svc->flags)) {
					BTERROR("Invalid flags at line %d", yylineno);
					return -1;
				}
			}
	| OPT_EXEC STRING 
			{ 
				strcpy(svc->cmd, $2);
				BTDEBUG("found exec: %s", svc->cmd);
			}
	| OPT_PROVIDER string
			{
				svc->prov = strdup($2);
			}
	| OPT_DESCRIPT string
			{
				svc->desc = strdup($2);
			}
	| OPT_COD params
			{
				str2cod_svc($2, &svc->cod);
				BTDEBUG("cod: %#.6x", svc->cod);
			}
	| OPT_SECURITY params
			{
				str2sec_level($2, &svc->security);
			}			
					
	//| error { /*yyclearin; yyerrok;*/ }
	;

/* ------------------------------------------------------------------- */

%%

void yyerror(char *s)
{
	BTERROR("reading config file: %s at line %d", s, yylineno);
}

void myyyerror(char *format, ...)
{
	int	n;
	char	str[255];
	va_list	ap;

	va_start(ap, format);
	n = vsnprintf(str, sizeof(str), format, ap);
	va_end(ap);

	yyerror(str);
}
	
	

int btsrv_read_config(char *file, int *service_num, int *device_num)
{
	int			err;
	extern FILE		*yyin;
	
	config_file = file;
	yyin = fopen(config_file, "r");
	if (!yyin) {
		BTERROR("Unable to open config file %s", config_file);
                return -1;
	}
	svcnum = 0;
	devnum = 0;
	err = yyparse();	/* parse config file */
	*service_num = svcnum;
	*device_num = devnum;
	fclose(yyin);
	yyin = NULL;

	return err;
}

