/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: request.c,v 1.24 2002/12/13 12:10:29 kds Exp $

   Contains the SDP service discovery client request handling code.

   Fixes:
   		Dmitry Kasatkin		- Continuation mechanism, MTU support
*/

#include <affix/config.h>

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include <affix/sdpsrv.h>
#include "request.h"
#include "servicedb.h"
#include "utils.h"
#include "attr.h"
#include "cstate.h"

extern void updateSDPServerDBTimestamp(void);
extern slist_t	*serviceList;

/*
 ** Generic data element sequence extractor. Builds
 ** a slist_t whose elements are those found in the 
 ** sequence. The data type of elements found in the
 ** sequence (now populated in slist_t) is returned
 ** in the reference pDataType
 */
int extractDataElementSequence(
		char *pduBuffer,
		slist_t **svcReqSeq,
		DataTypeDescriptor *pDataType,
		DataTypeDescriptor expectedType
		)
{
	int status, bytesScanned;
	char *pdata;
	DataTypeDescriptor seqType;
	int dataSize = 0;

	bytesScanned = 0; status = 0; 

	pdata = extractSequenceType(pduBuffer, &seqType, &dataSize, &bytesScanned);

	BTDEBUG("Seq type : %d", seqType);
	if ((pdata == NULL) ||
			((seqType != SequenceSize8) && (seqType != SequenceSize16))) {
		BTERROR("Unknown seq type");
		status = -1;
	}

	if (status != -1) {
		short numberOfElements;
		int seqLength;
		slist_t *pSeq;
		DataTypeDescriptor dataType;

		BTDEBUG("Data size : %d", dataSize);
		numberOfElements = 0; seqLength = 0; pSeq = NULL;
		for (;;) {
			char *pElem = NULL;
			int localSeqLength = 0;

			dataType = __get_u8(pdata);
			BTDEBUG("Data type : %d", dataType);

			switch (dataType) {
				case UnsignedInteger_16 :
					pdata += sizeof(DataTypeDescriptor);
					seqLength += sizeof(DataTypeDescriptor);
					pElem = (char *)malloc(sizeof(uint16_t));
					*(uint16_t *)pElem = ntohs(__get_u16(pdata));
					pdata += sizeof(uint16_t);
					seqLength += sizeof(uint16_t);
					break;

				case UnsignedInteger_32 :
					pdata += sizeof(DataTypeDescriptor);
					seqLength += sizeof(DataTypeDescriptor);
					pElem = (char *)malloc(sizeof(uint32_t));
					*(uint32_t *)pElem = ntohl(__get_u32(pdata));
					pdata += sizeof(uint32_t);
					seqLength += sizeof(uint32_t);
					break;

				case UUID_16 :
				case UUID_32 :
				case UUID_128 :
					pElem = (char *)malloc(sizeof(UUID));
					status = extractUUID(pdata, (UUID *)pElem, &localSeqLength);
					if (status == 0) {
						seqLength += localSeqLength;
						pdata += localSeqLength;
					}
					break;
			}

			if (status == 0) {
				s_list_append(&pSeq, pElem);
				numberOfElements++;
				BTDEBUG("No of elements : %d", numberOfElements);
				if (seqLength == dataSize)
					break;
				else
					if (seqLength > dataSize) {
						status = -1;
						break;
					}
			}
			else
				free(pElem);
		}

		if (status != -1) {
			*svcReqSeq = pSeq;
			bytesScanned += seqLength;
			*pDataType = dataType;
		}
	}

	if (status == 0)
		return bytesScanned;
	else
		return -1;
}



/*
 ** Service search request PDU. This methods extracts the "search pattern"
 ** which is a seqquence of UUIDs calls the matching function
 ** to find matching services
 */
int handleServiceSearchRequest(SDPRequest *requestPdu, PDUForm *pduForm)
{
	int status = 0, i;
	char *pdata;
	int bytesScanned = 0;
	slist_t *searchPattern;
	int maxExpectedServiceRecordCount, realServiceRecordCount;
	DataTypeDescriptor dataType;
	ContinuationState *pCState = NULL;
	char *pCacheBuffer = NULL;
	int handleSize = 0;
	long cStateId = -1;
	short responseCount = 0;
	short *pTotalRecordCount;
	short *pCurrentRecordCount;
	int	MTU;


	pdata = requestPdu->requestBuffer + sizeof(SdpPduHeader);

	searchPattern = NULL;
	bytesScanned = extractDataElementSequence(pdata, &searchPattern,
			&dataType, UUID_128);

	pdata += bytesScanned;
	maxExpectedServiceRecordCount = ntohs(__get_u16(pdata));
	BTDEBUG("Expected count: %d", maxExpectedServiceRecordCount);
	BTDEBUG("Bytes scanned : %d", bytesScanned);
	pdata += sizeof(uint16_t);

	/*
	 ** Check if continuation state exists, if yes attempt
	 ** to get response remainder from cache, else send error
	 */
	pCState = getContinuationState(pdata);

	MTU = getMTUFromFd(requestPdu->requestSocket);
	BTDEBUG("MTU: %d", MTU);
	MTU -= (sizeof(SdpPduHeader)+sizeof(uint16_t)+sizeof(uint16_t)+ContinuationStateSize);
	realServiceRecordCount = btmin(maxExpectedServiceRecordCount, (MTU >> 2));
	BTDEBUG("MTU: %d, realrec: %d", MTU, realServiceRecordCount);
	/*
	 ** Make space in the response buffer for total
	 ** and current record counts
	 */
	pdata = pduForm->data;

	/*
	 ** Total service record count = 0
	 */
	pTotalRecordCount = (short *)pdata;
	__put_u16(pdata, 0);
	pdata += sizeof(uint16_t);
	pduForm->dataSize += sizeof(uint16_t);

	/*
	 ** current service record count = 0
	 */
	pCurrentRecordCount = (short *)pdata;
	__put_u16(pdata, 0);
	pdata += sizeof(uint16_t);
	pduForm->dataSize += sizeof(uint16_t);

	if (pCState == NULL) {
		/*
		 ** For every element in the ServiceRecord DB,
		 ** do a pattern search
		 */
		int index = 0;
		
		handleSize = 0;

		for (;;) {
			void * data;
			ServiceRecord *svcRec;

			data = s_list_nth_data(serviceList, index);
			if (data != NULL) {
				svcRec = (ServiceRecord *)data;
				BTDEBUG("Checking svcRec : 0x%x", svcRec->serviceRecordHandle);
				if (matchUUID(searchPattern, svcRec->targetPattern)) {
					responseCount++; 
					__put_u32(pdata, htonl(svcRec->serviceRecordHandle));
					pdata += sizeof(ServiceRecordHandle);
					handleSize += sizeof(ServiceRecordHandle);
				}
				index++;
			} else
				break;
		}
		BTDEBUG("Match count : %d", responseCount);
		/*
		 ** Add "handleSize" to pduForm buffer size
		 */
		pduForm->dataSize += handleSize;
		__put_u16(pTotalRecordCount,  htons(responseCount));
		__put_u16(pCurrentRecordCount, htons(responseCount));

		if (responseCount > realServiceRecordCount) {
			/* 
			 ** We need to cache the response and generate a 
			 ** continuation state 
			 */
			cStateId = createContinuationStateBuffer(requestPdu->requestSocket, pduForm);
			/*
			 ** Subtract the handleSize since we now send only
			 ** a subset of handles
			 */
			pduForm->dataSize -= handleSize;
		} else {
			/*
			 ** NULL continuation state
			 */
			setContinuationStatePDU(pduForm, NULL);
		}
	}

	/*
	 ** Under both the conditions below, the response buffer is not
	 ** built up yet
	 */
	if ((pCState != NULL) || (cStateId != -1)) {
		short lastIndex = 0;

		if (pCState != NULL) {
			PDUForm *pCache;
			/*
			 ** Get the previous ContinuationState and obtain
			 ** the cached response
			 */
			pCache = getCachedResponse(requestPdu->requestSocket, pCState);
			if (pCache != NULL) {
				pCacheBuffer = pCache->data;
				/*
				 ** Get the responseCount from the cached buffer
				 */
				responseCount = ntohs(__get_u16(pCacheBuffer));

				/*
				 ** Get the index of the last ServiceRecord sent
				 */
				lastIndex = pCState->cStateValue.lastIndexSent;
			} else
				status = -1;
		} else {
			pCacheBuffer = pduForm->data;
			lastIndex = 0;
		}

		if (status == 0) {
			/*
			 ** Set the local buffer pointer to beyond the
			 ** current record count. And increment the cached
			 ** buffer pointer to beyond the counters ..
			 */
			pdata = (char *)pCurrentRecordCount + sizeof(uint16_t);

			/*
			 ** There is the toatalCount and the currentCount 
			 ** fields to increment beyond
			 */
			pCacheBuffer += (2 * sizeof(uint16_t));

			if (pCState == NULL) {
				handleSize = realServiceRecordCount << 2; //sizeof(ServiceRecordHandle); 
				i = realServiceRecordCount;
			} else {
				handleSize = 0;
				for (i = lastIndex; (i-lastIndex) < realServiceRecordCount &&
						i < responseCount; i++) {
					__put_u32(pdata, __get_u32(((uint32_t*)pCacheBuffer) + i));
					pdata += sizeof(ServiceRecordHandle);
					handleSize += sizeof(ServiceRecordHandle);
				}
			}

			/*
			 ** Add "handleSize" to pduForm buffer size
			 */
			pduForm->dataSize += handleSize;
			__put_u16(pTotalRecordCount,  htons(responseCount));
			__put_u16(pCurrentRecordCount, htons(i-lastIndex));

			if (i == responseCount) {
				/*
				 ** Set "null" continuationState & and remove it from cache
				 */
				setContinuationStatePDU(pduForm, NULL);
				removeCachedResponse(requestPdu->requestSocket, pCState);
			} else {
				/*
				 ** There is more to it .. set the lastIndexSent to
				 ** the new value and move on ..
				 */
				ContinuationState newState;

				BTDEBUG("Setting non-NULL CState");
				memset((char *)&newState, 0, sizeof(ContinuationState));
				if (pCState != NULL) {
					memcpy((char *)&newState, pCState, sizeof(ContinuationState));
					newState.cStateValue.lastIndexSent = i;
				} else {
					newState.timestamp = cStateId;
					newState.cStateValue.lastIndexSent = i;
				}
				setContinuationStatePDU(pduForm, &newState);
			}
		}
	}

	if (searchPattern != NULL)
		s_list_destroy(&searchPattern);

	return status;
}


/*
 ** This function generates the PDU form of a service record for
 ** a specified set of attributes and not the entire set. This is
 ** driven by a service discovery client which asks for attributes
 ** of a service record by speicifying a subset of attribute 
 ** identifiers
 */
void generateServiceRecordPduFormById(ServiceRecord *svcRec, ServiceAttributeIdentifier attrId,
					PDUForm *attrSubsetPduForm)
{
	SDPData *pData = NULL;

	pData = (SDPData *)getSDPData(svcRec, attrId);

	/*
	 ** If a subset pdu form of the service record is created
	 ** it is always in a local buffer "attrSubsetPduForm" defined
	 ** static to this file only
	 */
	if ((pData != NULL) && (pData->pduForm.data != NULL) && 
			(pData->pduForm.dataSize != 0))
		appendToPduForm(attrSubsetPduForm, &pData->pduForm);
}


/*
 ** Extract attribute identifiers from the request PDU.
 ** Clients could request a subset of attributes (by id)
 ** from a service record, instead of the whole set. The
 ** requested identifiers are present in the PDU form of
 ** the request
 */
int extractAttributesFromServiceRecord(
		char *requestBuffer,
		ServiceRecord *svcRec,
		slist_t *attrSeq,
		DataTypeDescriptor attrDataType,
		PDUForm *pduForm,
		int maxExpectedAttrSize
		)
{
	char *pdata;
	int status = 0;

	pdata = requestBuffer;
	/*
	 ** Set the maximum expected size before hand
	 */
	pduForm->bufferSize = maxExpectedAttrSize;
	if (svcRec != NULL) {
		/*
		 ** This is a request for the entire attribute range
		 */
		if (svcRec->pduForm.data == NULL) {
			BTDEBUG("Generating svc rec pdu form");
			generateServiceRecordPduForm(svcRec);
		}

#ifdef CONFIG_AFFIX_DEBUG
		if (attrSeq != NULL) {
			BTDEBUG("Entries in attr seq : %d", s_list_length(attrSeq));
		} else {
			BTDEBUG("NULL attribute descriptor");
		}
		BTDEBUG("AttrDataType : %d", attrDataType);
#endif

		if (attrSeq != NULL) {
			int attrLength;
			int index = 0;
			attrLength = 0;

			if (attrDataType == UnsignedInteger_16) {
				index = 0;
				for (;;) {
					void * pAttr;
					ServiceAttributeIdentifier attrId = 0;

					/*
					 ** Now we have a sequence of attribute identifiers for
					 ** which we need to generate PDU form
					 */
					pAttr = s_list_nth_data(attrSeq, index);
					if (pAttr == NULL) {
						BTDEBUG("Break out at index : %d", index);
						break;
					}
					index++;
					attrId = *(ServiceAttributeIdentifier *)pAttr;
					BTDEBUG("Attr id requested : 0x%x", attrId);
					generateServiceRecordPduFormById(svcRec, attrId, pduForm);
				}
			} else if (attrDataType == UnsignedInteger_32) {
					uint8_t completeAttrSetBuilt = 0;
					index = 0;

					for (;;) {
						void * pAttr = NULL;
						uint32_t attrRange = 0;
						ServiceAttributeIdentifier lowIdLim = 0;
						ServiceAttributeIdentifier highIdLim = 0;

						pAttr = s_list_nth_data(attrSeq, index);
						if (pAttr == NULL) {
							BTDEBUG("Break out at index : %d", index);
							break;
						}
						index++;
						attrRange = *(uint32_t *)pAttr;
						lowIdLim = ((0xffff0000 & attrRange) >> 16);
						highIdLim = 0x0000ffff & attrRange;

						BTDEBUG("attr range : 0x%x", attrRange);
						BTDEBUG("Low id : 0x%x", lowIdLim);
						BTDEBUG("High id : 0x%x", highIdLim);
						if ((lowIdLim == 0x0000) &&
								(highIdLim == 0xffff) &&
								(completeAttrSetBuilt == 0) &&
								(svcRec->pduForm.dataSize <= pduForm->bufferSize)) {
							completeAttrSetBuilt = 1;
							/*
							 ** Create a copy of it
							 */
							BTDEBUG("Copy attr pdu form : %d", svcRec->pduForm.dataSize);
							memcpy(pduForm->data, svcRec->pduForm.data, 
									svcRec->pduForm.dataSize);
							pduForm->dataSize = svcRec->pduForm.dataSize;
						} else {
							/*
							 ** Not the entire range of attributes, pick and choose
							 ** attributes requested
							 */
							int attrId = 0;
							for (attrId = lowIdLim; attrId <= highIdLim; attrId++) {
								generateServiceRecordPduFormById(svcRec, attrId, pduForm);
							}
						}
					}
				} else {
					status = InvalidRequestSyntax;
					BTERROR("Unexpected data type : 0x%x", attrDataType);
					BTERROR("Expect uint16_t or uint32_t");
				}
		} else {
			BTDEBUG("Attribute sequence is NULL");
		}
	} else {
		status = InvalidServiceRecordHandle;
	}

	return status;
}


/*
 ** A request for the attributes of a service record.
 ** First check if the service record (specified by
 ** service record handle) exists, then call the attribute
 ** streaming function
 */
int handleServiceAttributeRequest(SDPRequest *requestPdu, PDUForm *pduForm)
{
	char *pdata;
	int status = 0;
	ServiceRecord *svcRec;
	int maxResponseSize;
	ServiceRecordHandle svcRecHandle;
	ContinuationState *pCState = NULL;
	char *pResponse = NULL;
	short continuationStateSize = 0;
	slist_t *attrSeq = NULL;
	DataTypeDescriptor attrDataType = 0;
	int bytesScanned = 0;
	int	MTU;

	pdata = requestPdu->requestBuffer + sizeof(SdpPduHeader);
	svcRecHandle = ntohl(__get_u32(pdata));
	pdata += sizeof(ServiceRecordHandle);
	maxResponseSize = ntohs(__get_u16(pdata));
	pdata += sizeof(uint16_t);

	/*
	 ** Extract the attribute list
	 */
	bytesScanned = extractDataElementSequence(pdata, &attrSeq,
			&attrDataType, UnsignedInteger_16);
	pdata += bytesScanned;

	/*
	 ** Check if continuation state exists, if yes attempt
	 ** to get response remainder from cache, else send error
	 */
	pCState = getContinuationState(pdata);

	BTDEBUG("SvcRecHandle : 0x%x", svcRecHandle);
	BTDEBUG("maxResponseSize : %d", maxResponseSize);

	/* 
	 * get L2CAP mtu and calculate Attribute size
	 *
	 * We can send only (MTU - sizeof(SdpPduHeader) - sizeof(ContinuationState))
	 */
	MTU = getMTUFromFd(requestPdu->requestSocket);
	maxResponseSize = btmin(maxResponseSize,
			(MTU-sizeof(SdpPduHeader)-sizeof(ServiceRecordHandle)-
			 ContinuationStateSize-sizeof(uint16_t)));

	/*
	 * pull header for AttributeList byte count
	 */
	pduForm->data += sizeof(uint16_t);
	pduForm->bufferSize -= sizeof(uint16_t);

	if (pCState != NULL) {
		PDUForm *pCache = NULL;
		short bytesSent = 0;

		pCache = getCachedResponse(requestPdu->requestSocket, pCState);
		BTDEBUG("Obtained cached response : %p", pCache);
		if (pCache != NULL) {
			pResponse = pCache->data;
			bytesSent = btmin(maxResponseSize, 
					(pCache->dataSize - pCState->cStateValue.maxBytesSent));
			memcpy(pduForm->data, (pResponse + pCState->cStateValue.maxBytesSent), 
					bytesSent);
			pduForm->dataSize += bytesSent;
			pCState->cStateValue.maxBytesSent += bytesSent;
			BTDEBUG("Response size : %d sending now : %d bytes sent so far : %d", 
					pCache->dataSize, bytesSent, pCState->cStateValue.maxBytesSent);
			if (pCState->cStateValue.maxBytesSent == pCache->dataSize) {
				/* remove from cache */
				continuationStateSize = setContinuationStatePDU(pduForm, NULL);
				removeCachedResponse(requestPdu->requestSocket, pCState);
			} else
				continuationStateSize = setContinuationStatePDU(pduForm, pCState);
		} else {
			status = -1;
			BTERROR("NULL cache buffer and non-NULL continuation state");
		}
	} else {
		svcRec = findServiceRecord(svcRecHandle);
		status = extractAttributesFromServiceRecord(pdata, svcRec, attrSeq,
				attrDataType, pduForm, pduForm->bufferSize);
		//BTINFO("dataSize: %d, max: %d\n", pduForm->dataSize, maxResponseSize);
		if (pduForm->dataSize > maxResponseSize) {
			ContinuationState newState;

			memset((char *)&newState, 0, sizeof(ContinuationState));
			newState.timestamp = createContinuationStateBuffer(requestPdu->requestSocket, pduForm);
			/*
			 ** Reset the buffer size to the maximum expected and
			 ** set the ContinuationState
			 */
			BTDEBUG("Creating continuation state of size : %d", pduForm->dataSize);
			pduForm->dataSize = maxResponseSize;
			newState.cStateValue.maxBytesSent = maxResponseSize;
			continuationStateSize = setContinuationStatePDU(pduForm, &newState);
		} else
			continuationStateSize = setContinuationStatePDU(pduForm, NULL);
	}

	// push header
	pduForm->data -= sizeof(uint16_t);
	pduForm->bufferSize += sizeof(uint16_t);

	if (status != 0) {
		__put_u16(pduForm->data, htons(status));
		pduForm->dataSize = sizeof(ErrorCode);
	} else {
		/*
		 * set attribute list byte count
		 */
		__put_u16(pduForm->data, htons(pduForm->dataSize - continuationStateSize));
		pduForm->dataSize += sizeof(uint16_t);
	}

	return status;
}


/*
 ** The single request for a combined "service search" and
 ** "attribute extraction"
 */
int handleServiceSearchAttributeRequest(SDPRequest *requestPdu, PDUForm *pduForm)
{
	int status = 0;
	char *pdata;
	int bytesScanned = 0;
	slist_t *searchPattern;
	int maxExpectedAttrSize;
	DataTypeDescriptor dataType;
	ContinuationState *pCState = NULL;
	char *pResponse = NULL;
	short continuationStateSize = 0;
	slist_t *attrSeq = NULL;
	DataTypeDescriptor attrDataType = 0;
	int index = 0;
	int responseCount = 0;
	PDUForm localPduForm;
	int	MTU;

	BTDEBUG("In svc srch attr request");
	pdata = requestPdu->requestBuffer + sizeof(SdpPduHeader);

	searchPattern = NULL;
	bytesScanned = extractDataElementSequence(pdata, &searchPattern,
			&dataType, UnsignedInteger_16);
	BTDEBUG("Bytes scanned : %d", bytesScanned);

	pdata += bytesScanned;
	maxExpectedAttrSize = ntohs(__get_u16(pdata));
	pdata += sizeof(uint16_t);
	BTDEBUG("Max Attr expected : %d", maxExpectedAttrSize);

	/*
	 ** Extract the attribute list
	 */
	bytesScanned = extractDataElementSequence(pdata, &attrSeq,
			&attrDataType, UnsignedInteger_16);
	pdata += bytesScanned;

	/*
	 ** Check if continuation state exists, if yes attempt
	 ** to get response remainder from cache, else send error
	 */
	pCState = getContinuationState(pdata);	//this is my Continuation Information

	localPduForm.data = (char *)malloc(USHRT_MAX);
	localPduForm.dataSize = 0;
	localPduForm.bufferSize = USHRT_MAX;

	/* 
	 * get L2CAP mtu and calculate Attribute size
	 *
	 * We can send only (MTU - sizeof(SdpPduHeader) - sizeof(ContinuationState))
	 */
	MTU = getMTUFromFd(requestPdu->requestSocket);
	maxExpectedAttrSize = btmin(maxExpectedAttrSize,
			(MTU-sizeof(SdpPduHeader)-ContinuationStateSize-sizeof(uint16_t)));
	//maxExpectedAttrSize = 128;

	/*
	 * pull header for AttributeList byte count
	 */
	pduForm->data += sizeof(uint16_t);
	pduForm->bufferSize -= sizeof(uint16_t);

	if (pCState == NULL) {
		/*
		 * No Continuation State -> Create New Response
		 */
		for (;;) {
			void * data;
			ServiceRecord *svcRec;

			localPduForm.dataSize = 0;
			memset(localPduForm.data, 0, USHRT_MAX);

			data = s_list_nth_data(serviceList, index);
			if (data != NULL) {
				svcRec = (ServiceRecord *)data;
				if (matchUUID(searchPattern, svcRec->targetPattern)) {
					responseCount++;
					status = extractAttributesFromServiceRecord(pdata, svcRec, attrSeq,
							attrDataType, &localPduForm, localPduForm.bufferSize);
					BTDEBUG("Response count : %d", responseCount);
					BTDEBUG("Local PDU size : %d", localPduForm.dataSize);
					if (status == -1) {
						BTDEBUG("Extract attr from svc rec returns err");
						break;
					} else {
						if (pduForm->dataSize + localPduForm.dataSize < pduForm->bufferSize) {
							// to be sure no relocations
							appendToPduForm(pduForm, &localPduForm);
						} else {
							BTERROR("Relocation needed");
							break;
						}
						BTDEBUG("Net PDU size : %d", pduForm->dataSize);
					}
				}
				index++;
			} else 
				break;
		}
		if (pduForm->dataSize > maxExpectedAttrSize) {
			ContinuationState newState;

			memset((char *)&newState, 0, sizeof(ContinuationState));
			newState.timestamp = createContinuationStateBuffer(requestPdu->requestSocket, pduForm);
			/*
			 ** Reset the buffer size to the maximum expected and
			 ** set the ContinuationState
			 */
			pduForm->dataSize = maxExpectedAttrSize;
			newState.cStateValue.maxBytesSent = maxExpectedAttrSize;
			continuationStateSize = setContinuationStatePDU(pduForm, &newState);
		} else
			continuationStateSize = setContinuationStatePDU(pduForm, NULL);
	} else {
		/*
		 * Continuation State exists -> get from cache
		 */
		PDUForm *pCache;
		uint16_t bytesSent = 0;

		/*
		 ** Get the cached response from the buffer
		 */
		pCache = getCachedResponse(requestPdu->requestSocket, pCState);
		if (pCache != NULL) {
			pResponse = pCache->data;
			bytesSent = btmin(maxExpectedAttrSize, 
					(pCache->dataSize - pCState->cStateValue.maxBytesSent));
			memcpy(pduForm->data, (pResponse + pCState->cStateValue.maxBytesSent), 
					bytesSent);
			pduForm->dataSize += bytesSent;
			pCState->cStateValue.maxBytesSent += bytesSent;
			if (pCState->cStateValue.maxBytesSent == pCache->dataSize) {
				/* remove response from cache as well */
				continuationStateSize = setContinuationStatePDU(pduForm, NULL);
				removeCachedResponse(requestPdu->requestSocket, pCState);
			} else
				continuationStateSize = setContinuationStatePDU(pduForm, pCState);
		} else {
			status = -1;
			BTDEBUG("Non null continuation state, but null cache buffer");
		}
	}

	if ((responseCount == 0) && (pCState == NULL)) {
		//we've not found anything
		appendToPduForm(pduForm, &localPduForm);
	}

	// push header
	pduForm->data -= sizeof(uint16_t);
	pduForm->bufferSize += sizeof(uint16_t);

	if (status != 0) {
		__put_u16(pduForm->data, htons(status));
		pduForm->dataSize = sizeof(ErrorCode);
	} else {
		/*
		 * set attribute list byte count
		 */
		__put_u16(pduForm->data, htons(pduForm->dataSize - continuationStateSize));
		pduForm->dataSize += sizeof(uint16_t);
	}

	free(localPduForm.data);
	s_list_destroy(&searchPattern);
	return status;
}


ServiceRecord *serverExtractPDU(char *pdata, ServiceRecordHandle handleExpected, int *bytesScanned)
{
	int localExtractedLength = 0;
	char *localSeqPtr = NULL;
	DataTypeDescriptor dtd;
	int sequenceLength = 0;
	ServiceRecord *svcRec = NULL;
	ServiceAttributeIdentifier lookAheadAttrId = 0xffff;

	localSeqPtr = extractSequenceType(pdata, &dtd, &sequenceLength, 
			bytesScanned);
	lookAheadAttrId = ntohs(__get_u16(localSeqPtr +	sizeof(DataTypeDescriptor)));
	BTDEBUG("Look ahead attr id : %d", lookAheadAttrId);

	/*
	 ** Check if the PDU already has a service record handle,
	 ** int which case this may be a update request
	 */
	if (handleExpected != 0xffffffff)
		svcRec = findServiceRecord(handleExpected);
	
	if (svcRec == NULL) {
		svcRec = createServiceRecord();
		svcRec->attributeList = NULL;
	}

	localExtractedLength = extractAttributes(svcRec, localSeqPtr, sequenceLength);
	if (localExtractedLength >= 0)
		*bytesScanned += localExtractedLength;

	return svcRec;
}

/*
 ** Request to create a brand new service record
 ** The newly created service record is added to the
 ** service repository
 */
int handleServiceRegistrationRequest(SDPRequest *sdpReq, PDUForm *response)
{
	int status = 0;
	char *pdata;
	int bytesScanned = 0;
	ServiceRecord *svcRec;
	SdpPduHeader *sdpPduHeader;
	int length = 0;
	SDPData *browseAttrData = NULL;
	SDPData *pData;


	/*
	 ** Extract buffer contents
	 */
	sdpPduHeader = (SdpPduHeader *)sdpReq->requestBuffer;
	pdata = sdpReq->requestBuffer + sizeof(SdpPduHeader);

	/*
	 ** Extract and save image of PDU, since we need
	 ** it when clients request thsi attribute
	 */
	svcRec = serverExtractPDU(pdata, 0xffffffff, &bytesScanned);
	if (svcRec != NULL) {
		svcRec->serviceRecordHandle = (ServiceRecordHandle)svcRec;
		addServiceRecord(svcRec);
		pData = createSDPData(UnsignedInteger_32, (char *)&svcRec->serviceRecordHandle);
		addToAttributeList(svcRec, AttrServiceRecordHandle, pData, 1);
		/*
		 ** Also check if the browse group descriptor is NULL,
		 ** ensure that the record belongs to the ROOT group
		 */
		if ((browseAttrData = getSDPData(svcRec, AttrBrowseGroupList)) == NULL)
			addUUID16ToTargetPattern(svcRec, PublicBrowseGroup);
	} else {
		status = -1;
	}

	if (status == 0) {
		generateServiceRecordPduForm(svcRec);
		updateSDPServerDBTimestamp();

		/*
		 ** Build a response buffer
		 */
		pdata = response->data;
		__put_u32(pdata, htonl(svcRec->serviceRecordHandle));
		BTDEBUG("Sending back svcRecHandle : 0x%x", svcRec->serviceRecordHandle);
		length += sizeof(uint32_t);
		response->dataSize = length;
		
		/* set owner of the service */
		svcRec->ownerSocket = sdpReq->requestSocket;
	} else {
		__put_u16(response->data, htons(InvalidRequestSyntax));
		response->dataSize = sizeof(ErrorCode);
	}

	return status;
}


/*
 ** Register a service record (this means updating the
 ** contents of a newly created service record or updating
 ** the contents of an existing record)
 */
int handleServiceUpdateRequest(SDPRequest *sdpReq, PDUForm *response)
{
	char *pdata;
	int status = 0;
	ServiceRecordHandle svcRecHandle;
	ServiceRecord *svcRecNew = NULL;
	ServiceRecord *svcRecOld = NULL;
	SdpPduHeader *sdpPduHeader;
	int bytesScanned = 0;

	/*
	 ** Extract service record handle
	 */
	sdpPduHeader = (SdpPduHeader *)sdpReq->requestBuffer;
	pdata = sdpReq->requestBuffer + sizeof(SdpPduHeader);
	svcRecHandle = ntohl(__get_u32(pdata));
	BTDEBUG("Svc Rec Handle : 0x%x\n", svcRecHandle);
	pdata += sizeof(ServiceRecordHandle);

	svcRecOld = findServiceRecord(svcRecHandle);
	BTDEBUG("SvcRecOld : 0x%x\n", (uint32_t)svcRecOld);
	if (svcRecOld != NULL) {
		svcRecNew = serverExtractPDU(pdata, svcRecHandle, &bytesScanned);
		if ((svcRecNew != NULL) && 
				(svcRecHandle == svcRecNew->serviceRecordHandle)) {
			generateServiceRecordPduForm(svcRecNew);
			updateSDPServerDBTimestamp();
		} else {
			BTDEBUG("SvcRecHandle : 0x%x\n", svcRecHandle);
			BTDEBUG("SvcRecHandleNew : 0x%x\n", svcRecNew->serviceRecordHandle);
			BTDEBUG("SvcRecNew : 0x%x\n", (uint32_t)svcRecNew);
			BTDEBUG("SvcRecOld : 0x%x\n", (uint32_t)svcRecOld);
			BTDEBUG("Failure to update, restore old value\n");
			if (svcRecNew != NULL)
				freeServiceRecord(svcRecNew);
			status = InvalidRequestSyntax;
		}
	}
	else
		status = InvalidServiceRecordHandle;

	pdata = response->data;
	__put_u16(pdata, htons(status));
	response->dataSize = sizeof(ErrorCode);
	return status;
}


/*
 ** Remove a registered service record
 */
int handleServiceRemovalRequest(SDPRequest *sdpReq, PDUForm *response)
{
	char *pdata;
	int status = 0;
	ServiceRecordHandle svcRecHandle;
	ServiceRecord *svcRec;

	/*
	 ** Extract service record handle
	 */
	pdata = sdpReq->requestBuffer + sizeof(SdpPduHeader);
	svcRecHandle = ntohl(__get_u32(pdata));
	pdata += sizeof(ServiceRecordHandle);

	svcRec = findServiceRecord(svcRecHandle);
	if (svcRec != NULL) {
		status = removeServiceRecord(svcRecHandle);
		freeServiceRecord(svcRec);
		if (status == 0)
			updateSDPServerDBTimestamp();
	}
	else {
		status = InvalidServiceRecordHandle;
		BTDEBUG("Could not find record : 0x%x\n", svcRecHandle);
	}

	pdata = response->data;
	__put_u16(pdata, htons(status));
	response->dataSize = sizeof(ErrorCode);

	return status;
}


/*
 ** Top level request processor. Calls the appropriate processing
 ** function based on request type. Handles service registration
 ** client requests also
 */
void processRequest(SDPRequest *request)
{
	char *pdata;
	int status = 0;
	SdpPduHeader *requestPduHeader;
	SdpPduHeader *responsePduHeader;
	PDUForm responsePdu;
	char *bufferStart;
	int bytesSent = 0;

	BTDEBUG("In process req");
	requestPduHeader = (SdpPduHeader *)request->requestBuffer;

	bufferStart = (char *)malloc(USHRT_MAX);
	memset((void *)bufferStart, 0, USHRT_MAX);
	responsePdu.data = bufferStart + sizeof(SdpPduHeader);
	responsePdu.dataSize = 0;
	responsePdu.bufferSize = USHRT_MAX - sizeof(SdpPduHeader);
	responsePduHeader = (SdpPduHeader *)bufferStart;

	switch (requestPduHeader->pduId) {
		case SDP_ServiceSearchRequest :
			BTDEBUG("Got a svc srch req");
			status = handleServiceSearchRequest(request, &responsePdu);

			responsePduHeader->pduId =
				((status == 0) ? SDP_ServiceSearchResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceAttributeRequest :
			BTDEBUG("Got a svc attr req");
			status = handleServiceAttributeRequest(request, &responsePdu);

			responsePduHeader->pduId =
				((status == 0) ? SDP_ServiceAttributeResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceSearchAttributeRequest :
			BTDEBUG("Got a svc srch attr req");
			status = handleServiceSearchAttributeRequest(request, &responsePdu);

			responsePduHeader->pduId = ((status == 0) ? 
					SDP_ServiceSearchAttributeResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceRegistrationRequest :
			BTDEBUG("Service register request");
			status = handleServiceRegistrationRequest(request, &responsePdu);

			responsePduHeader->pduId = ((status == 0) ? 
					SDP_ServiceRegistrationResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceUpdateRequest :
			BTDEBUG("Service update request");
			status = handleServiceUpdateRequest(request, &responsePdu);

			responsePduHeader->pduId = ((status == 0) ? 
					SDP_ServiceUpdateResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceRemovalRequest :
			BTDEBUG("Service removal request");
			status = handleServiceRemovalRequest(request, &responsePdu);

			responsePduHeader->pduId = ((status == 0) ? 
					SDP_ServiceRemovalResponse : SDP_ErrorResponse);
			break;

		default :
			BTERROR("Unknown PDU ID : 0x%x received", requestPduHeader->pduId);
			pdata = responsePdu.data;
			__put_u16(pdata, htons(InvalidRequestSyntax));
			responsePdu.dataSize = sizeof(ErrorCode);
			responsePduHeader->pduId = SDP_ErrorResponse;
			break;
	}

	BTDEBUG("Status : %d", status);
	BTDEBUG("Sending a response back");
	responsePduHeader->transactionId = requestPduHeader->transactionId;
	responsePduHeader->paramLength = htons(responsePdu.dataSize);
	/*
	 ** Now point back to the real buffer start and set the
	 ** real response length
	 */
	responsePdu.dataSize += sizeof(SdpPduHeader);
	responsePdu.data = bufferStart;

	/*
	 ** Stream the response PDU
	 */
	bytesSent = send(request->requestSocket, responsePdu.data, 
			responsePdu.dataSize, 0);

	BTDEBUG("Bytes Sent : %d, requested: %d", bytesSent, responsePdu.dataSize);

	free(responsePdu.data);
	free(request->requestBuffer);
	free(request);
}

/*
 * Remove services by RequestSocket
 *
 */
int handleServiceDisconnectRequest(int fd)
{
	int	i, status;

start_again:
	for (i = 0; i < s_list_length(serviceList); i++) {
		ServiceRecord *svcRec;

		svcRec = (ServiceRecord *)s_list_nth_data(serviceList, i);
		BTDEBUG("fd: %d, svc->fd: %d\n", fd, svcRec->ownerSocket);
		if (svcRec->ownerSocket == fd) {
			status = removeServiceRecord(svcRec->serviceRecordHandle);
			freeServiceRecord(svcRec);
			goto start_again;
		}
	}
	return 0;	
}

